# All effects and their IDs
none = 0
"""Attributes for the **none** effect are: \n
... none... Just like Conditions... People these days... """
change_diplomacy = 1
"""Attributes for the **change_diplomacy** effect are: \n
- diplomacy
- player_source
- player_target"""
research_technology = 2
"""Attributes for the **research_technology** effect are: \n
- player_source
- technology
- force_research_technology"""
send_chat = 3
"""Attributes for the **send_chat** effect are: \n
- player_source
- string_id
- message
- sound_name"""
play_sound = 4
"""Attributes for the **play_sound** effect are: \n
- player_source
- location_x
- location_y
- sound_name"""
tribute = 5
"""Attributes for the **tribute** effect are: \n
- quantity
- tribute_list
- player_source
- player_target"""
unlock_gate = 6
"""Attributes for the **unlock_gate** effect are: \n
- number_of_units_selected
- selected_object_id"""
lock_gate = 7
"""Attributes for the **lock_gate** effect are: \n
- number_of_units_selected
- selected_object_id"""
activate_trigger = 8
"""Attributes for the **activate_trigger** effect are: \n
- trigger_id"""
deactivate_trigger = 9
"""Attributes for the **deactivate_trigger** effect are: \n
- trigger_id"""
create_object = 11
"""Attributes for the **create_object** effect are: \n
- object_list_unit_id
- player_source
- location_x
- location_y
- item_id
- facet"""
task_object = 12
"""Attributes for the **task_object** effect are: \n
- number_of_units_selected
- object_list_unit_id
- player_source
- location_x
- location_y
- area_1_x
- area_1_y
- area_2_x
- area_2_y
- object_group
- object_type
- selected_object_id"""
declare_victory = 13
"""Attributes for the **declare_victory** effect are: \n
- player_source
- enabled_or_victory"""
kill_object = 14
"""Attributes for the **kill_object** effect are: \n
- number_of_units_selected
- object_list_unit_id
- player_source
- area_1_x
- area_1_y
- area_2_x
- area_2_y
- object_group
- object_type
- selected_object_id"""
remove_object = 15
"""Attributes for the **remove_object** effect are: \n
- number_of_units_selected
- object_list_unit_id
- player_source
- area_1_x
- area_1_y
- area_2_x
- area_2_y
- object_group
- object_type
- selected_object_id"""
change_view = 16
"""Attributes for the **change_view** effect are: \n
- player_source
- location_x
- location_y
- scroll"""
unload = 17
"""Attributes for the **unload** effect are: \n
- number_of_units_selected
- object_list_unit_id
- player_source
- location_x
- location_y
- area_1_x
- area_1_y
- area_2_x
- area_2_y
- object_group
- object_type
- selected_object_id"""
change_ownership = 18
"""Attributes for the **change_ownership** effect are: \n
- number_of_units_selected
- object_list_unit_id
- player_source
- player_target
- area_1_x
- area_1_y
- area_2_x
- area_2_y
- object_group
- object_type
- flash_object
- selected_object_id"""
patrol = 19
"""Attributes for the **patrol** effect are: \n
- number_of_units_selected
- object_list_unit_id
- player_source
- location_x
- location_y
- area_1_x
- area_1_y
- area_2_x
- area_2_y
- object_group
- object_type
- selected_object_id"""
display_instructions = 20
"""Attributes for the **display_instructions** effect are: \n
- object_list_unit_id
- player_source
- string_id
- display_time
- instruction_panel_position
- play_sound
- message
- sound_name"""
clear_instructions = 21
"""Attributes for the **clear_instructions** effect are: \n
- instruction_panel_position"""
freeze_object = 22
"""Attributes for the **freeze_object** effect are: \n
- number_of_units_selected
- object_list_unit_id
- player_source
- area_1_x
- area_1_y
- area_2_x
- area_2_y
- object_group
- object_type
- selected_object_id"""
use_advanced_buttons = 23
"""Attributes for the **use_advanced_buttons** effect are: \n
None. \n
Please don't use this effect. Please."""
damage_object = 24
"""Attributes for the **damage_object** effect are: \n
- quantity
- number_of_units_selected
- object_list_unit_id
- player_source
- area_1_x
- area_1_y
- area_2_x
- area_2_y
- object_group
- object_type
- selected_object_id"""
place_foundation = 25
"""Attributes for the **place_foundation** effect are: \n
- object_list_unit_id
- player_source
- location_x
- location_y"""
change_object_name = 26
"""Attributes for the **change_object_name** effect are: \n
- number_of_units_selected
- object_list_unit_id
- player_source
- string_id
- area_1_x
- area_1_y
- area_2_x
- area_2_y
- message
- selected_object_id"""
change_object_hp = 27
"""Attributes for the **change_object_hp** effect are: \n
- quantity
- number_of_units_selected
- object_list_unit_id
- player_source
- area_1_x
- area_1_y
- area_2_x
- area_2_y
- object_group
- object_type
- operation
- selected_object_id"""
change_object_attack = 28
"""Attributes for the **change_object_attack** effect are: \n
- aa_quantity
- aa_armor_or_attack_type
- number_of_units_selected
- object_list_unit_id
- player_source
- area_1_x
- area_1_y
- area_2_x
- area_2_y
- object_group
- object_type
- operation
- selected_object_id"""
stop_object = 29
"""Attributes for the **stop_object** effect are: \n
- number_of_units_selected
- object_list_unit_id
- player_source
- area_1_x
- area_1_y
- area_2_x
- area_2_y
- object_group
- object_type
- selected_object_id"""
attack_move = 30
"""Attributes for the **attack_move** effect are: \n
- number_of_units_selected
- object_list_unit_id
- player_source
- location_x
- location_y
- area_1_x
- area_1_y
- area_2_x
- area_2_y
- object_group
- object_type
- selected_object_id"""
change_object_armor = 31
"""Attributes for the **change_object_armor** effect are: \n
- aa_quantity
- aa_armor_or_attack_type
- number_of_units_selected
- object_list_unit_id
- player_source
- area_1_x
- area_1_y
- area_2_x
- area_2_y
- object_group
- object_type
- operation
- selected_object_id"""
change_object_range = 32
"""Attributes for the **change_object_range** effect are: \n
- quantity
- number_of_units_selected
- object_list_unit_id
- player_source
- area_1_x
- area_1_y
- area_2_x
- area_2_y
- object_group
- object_type
- operation
- selected_object_id"""
change_object_speed = 33
"""Attributes for the **change_object_speed** effect are: \n
- quantity
- number_of_units_selected
- object_list_unit_id
- player_source
- area_1_x
- area_1_y
- area_2_x
- area_2_y
- object_group
- object_type
- selected_object_id"""
heal_object = 34
"""Attributes for the **heal_object** effect are: \n
- quantity
- number_of_units_selected
- object_list_unit_id
- player_source
- area_1_x
- area_1_y
- area_2_x
- area_2_y
- object_group
- object_type
- selected_object_id"""
teleport_object = 35
"""Attributes for the **teleport_object** effect are: \n
- number_of_units_selected
- object_list_unit_id
- player_source
- location_x
- location_y
- area_1_x
- area_1_y
- area_2_x
- area_2_y
- object_group
- object_type
- selected_object_id"""
change_object_stance = 36
"""Attributes for the **change_object_stance** effect are: \n
- number_of_units_selected
- object_list_unit_id
- player_source
- area_1_x
- area_1_y
- area_2_x
- area_2_y
- object_group
- object_type
- attack_stance
- selected_object_id"""
display_timer = 37
"""Attributes for the **display_timer** effect are: \n
- string_id
- display_time
- time_unit
- variable_or_timer
- message"""
enable_disable_object = 38
"""Attributes for the **enable_disable_object** effect are: \n
- object_list_unit_id
- player_source
- enabled_or_victory
- item_id"""
enable_disable_technology = 39
"""Attributes for the **enable_disable_technology** effect are: \n
- player_source
- technology
- enabled_or_victory
- item_id"""
change_object_cost = 40
"""Attributes for the **change_object_cost** effect are: \n
- object_list_unit_id
- player_source
- food
- wood
- stone
- gold"""
set_player_visibility = 41
"""Attributes for the **set_player_visibility** effect are: \n
- player_source
- player_target
- visibility_state"""
change_object_icon = 42
"""Attributes for the **change_object_icon** effect are: \n
- number_of_units_selected
- object_list_unit_id
- player_source
- area_1_x
- area_1_y
- area_2_x
- area_2_y
- object_group
- object_type
- object_list_unit_id_2
- selected_object_id"""
replace_object = 43
"""Attributes for the **replace_object** effect are: \n
- number_of_units_selected
- object_list_unit_id
- player_source
- player_target
- area_1_x
- area_1_y
- area_2_x
- area_2_y
- object_group
- object_type
- object_list_unit_id_2
- selected_object_id"""
change_object_description = 44
"""Attributes for the **change_object_description** effect are: \n
- object_list_unit_id
- player_source
- string_id
- message"""
change_player_name = 45
"""Attributes for the **change_player_name** effect are: \n
- player_source
- string_id
- message"""
change_train_location = 46
"""Attributes for the **change_train_location** effect are: \n
- object_list_unit_id
- player_source
- object_list_unit_id_2
- button_location"""
change_research_location = 47
"""Attributes for the **change_research_location** effect are: \n
- player_source
- technology
- object_list_unit_id_2
- button_location"""
change_civilization_name = 48
"""Attributes for the **change_civilization_name** effect are: \n
- player_source
- string_id
- message"""
create_garrisoned_object = 49
"""Attributes for the **create_garrisoned_object** effect are: \n
- number_of_units_selected
- object_list_unit_id
- player_source
- area_1_x
- area_1_y
- area_2_x
- area_2_y
- object_list_unit_id_2
- selected_object_id"""
acknowledge_ai_signal = 50
"""Attributes for the **acknowledge_ai_signal** effect are: \n
- ai_signal_value"""
modify_attribute = 51
"""Attributes for the **modify_attribute** effect are: \n
- quantity
- object_list_unit_id
- player_source
- item_id
- operation
- object_attributes"""
modify_resource = 52
"""Attributes for the **modify_resource** effect are: \n
- quantity
- tribute_list
- player_source
- item_id
- operation"""
modify_resource_by_variable = 53
"""Attributes for the **modify_resource_by_variable** effect are: \n
- tribute_list
- player_source
- item_id
- operation
- from_variable"""
change_variable = 56
"""Attributes for the **change_variable** effect are: \n
- quantity
- operation
- from_variable
- message"""
clear_timer = 57
"""Attributes for the **clear_timer** effect are: \n
- variable_or_timer"""

attribute_naming_conversion = {
    "effect_type": "Effect type",
    "ai_script_goal": "AI script goal",
    "quantity": "Quantity",
    "aa_armor_or_attack_type": "AA Armor/Attack Type",
    "aa_quantity": "AA Quantity",
    "tribute_list": "Tribute List",
    "diplomacy": "Diplomacy",
    "number_of_units_selected": "Number of units selected",
    "object_list_unit_id": "Object list unit ID",
    "player_source": "Player Source",
    "player_target": "Player Target",
    "technology": "Technology",
    "string_id": "String ID",
    "display_time": "Display Time (display instructions)",
    "trigger_id": "Trigger ID (activate/deactivate)",
    "location_x": "Location X",
    "location_y": "Location Y",
    "area_1_x": "Area 1 X",
    "area_1_y": "Area 1 Y",
    "area_2_x": "Area 2 X",
    "area_2_y": "Area 2 Y",
    "object_group": "Object Group",
    "object_type": "Object Type",
    "instruction_panel_position": "Instruction Panel Position",
    "attack_stance": "Attack Stance",
    "time_unit": "Time unit (second, minutes, years)",
    "enabled_or_victory": "Enabled/Victory",
    "food": "Food",
    "wood": "Wood",
    "stone": "Stone",
    "gold": "Gold",
    "item_id": "Item ID",
    "flash_object": "Flash Object",
    "force_research_technology": "Force Research Technology",
    "visibility_state": "Visibility State",
    "scroll": "Scroll (Set view)",
    "operation": "Operation",
    "object_list_unit_id_2": "Object list unit ID 2",
    "button_location": "Button Location",
    "ai_signal_value": "AI signal Value",
    "object_attributes": "Object attributes",
    "from_variable": "From Variable",
    "variable_or_timer": "Variable/Timer",
    "facet": "Facet",
    "play_sound": "Play Sound",
    "message": "Message",
    "sound_name": "Sound (event) name",
    "selected_object_id": "Selected Object(s) ID",
}

effect_identifier_conversion = {
    0: "none",
    1: "change_diplomacy",
    2: "research_technology",
    3: "send_chat",
    4: "play_sound",
    5: "tribute",
    6: "unlock_gate",
    7: "lock_gate",
    8: "activate_trigger",
    9: "deactivate_trigger",
    11: "create_object",
    12: "task_object",
    13: "declare_victory",
    14: "kill_object",
    15: "remove_object",
    16: "change_view",
    17: "unload",
    18: "change_ownership",
    19: "patrol",
    20: "display_instructions",
    21: "clear_instructions",
    22: "freeze_object",
    23: "use_advanced_buttons",
    24: "damage_object",
    25: "place_foundation",
    26: "change_object_name",
    27: "change_object_hp",
    28: "change_object_attack",
    29: "stop_object",
    30: "attack_move",
    31: "change_object_armor",
    32: "change_object_range",
    33: "change_object_speed",
    34: "heal_object",
    35: "teleport_object",
    36: "change_object_stance",
    37: "display_timer",
    38: "enable/disable_object",
    39: "enable/disable_technology",
    40: "change_object_cost",
    41: "set_player_visibility",
    42: "change_object_icon",
    43: "replace_object",
    44: "change_object_description",
    45: "change_player_name",
    46: "change_train_location",
    47: "change_research_location",
    48: "change_civilization_name",
    49: "create_garrisoned_object",
    50: "acknowledge_ai_signal",
    51: "modify_attribute",
    52: "modify_resource",
    53: "modify_resource_by_variable",
    56: "change_variable",
    57: "clear_timer",
}

# Credits: Eti JS @ https://stackoverflow.com/a/58746861/7230293
attribute_naming_conversion.update(dict(map(reversed, attribute_naming_conversion.items())))
effect_identifier_conversion.update(dict(map(reversed, effect_identifier_conversion.items())))


empty_attributes = {
    "effect_type": -1,
    "ai_script_goal": -1,
    "aa_quantity": -1,
    "aa_armor_or_attack_type": -1,
    "quantity": -1,
    "tribute_list": -1,
    "diplomacy": -1,
    "number_of_units_selected": -1,
    "object_list_unit_id": -1,
    "player_source": -1,
    "player_target": -1,
    "technology": -1,
    "string_id": -1,
    "display_time": -1,
    "trigger_id": -1,
    "location_x": -1,
    "location_y": -1,
    "area_1_x": -1,
    "area_1_y": -1,
    "area_2_x": -1,
    "area_2_y": -1,
    "object_group": -1,
    "object_type": -1,
    "instruction_panel_position": -1,
    "attack_stance": -1,
    "time_unit": -1,
    "enabled_or_victory": -1,
    "food": -1,
    "wood": -1,
    "stone": -1,
    "gold": -1,
    "item_id": -1,
    "flash_object": -1,
    "force_research_technology": -1,
    "visibility_state": -1,
    "scroll": -1,
    "operation": -1,
    "object_list_unit_id_2": -1,
    "button_location": -1,
    "ai_signal_value": -1,
    "object_attributes": -1,
    "from_variable": -1,
    "variable_or_timer": -1,
    "facet": -1,
    "play_sound": -1,
    "message": '',
    "sound_name": '',
    "selected_object_id": -1,
}

attributes = {
    0: [
        "effect_type",
    ],
    50: [
        "effect_type",
        "ai_signal_value",
    ],
    8: [
        "effect_type",
        "trigger_id",
    ],
    30: [
        "effect_type",
        "number_of_units_selected",
        "object_list_unit_id",
        "player_source",
        "location_x",
        "location_y",
        "area_1_x",
        "area_1_y",
        "area_2_x",
        "area_2_y",
        "object_group",
        "object_type",
        "selected_object_id",
    ],
    48: [
        "effect_type",
        "player_source",
        "string_id",
        "message",
    ],
    1: [
        "effect_type",
        "diplomacy",
        "player_source",
        "player_target",
    ],
    31: [
        "effect_type",
        "aa_quantity",
        "aa_armor_or_attack_type",
        "number_of_units_selected",
        "object_list_unit_id",
        "player_source",
        "area_1_x",
        "area_1_y",
        "area_2_x",
        "area_2_y",
        "object_group",
        "object_type",
        "operation",
        "selected_object_id",
    ],
    28: [
        "effect_type",
        "aa_quantity",
        "aa_armor_or_attack_type",
        "number_of_units_selected",
        "object_list_unit_id",
        "player_source",
        "area_1_x",
        "area_1_y",
        "area_2_x",
        "area_2_y",
        "object_group",
        "object_type",
        "operation",
        "selected_object_id",
    ],
    40: [
        "effect_type",
        "object_list_unit_id",
        "player_source",
        "food",
        "wood",
        "stone",
        "gold",
    ],
    44: [
        "effect_type",
        "object_list_unit_id",
        "player_source",
        "string_id",
        "message",
    ],
    27: [
        "effect_type",
        "quantity",
        "number_of_units_selected",
        "object_list_unit_id",
        "player_source",
        "area_1_x",
        "area_1_y",
        "area_2_x",
        "area_2_y",
        "object_group",
        "object_type",
        "operation",
        "selected_object_id",
    ],
    42: [
        "effect_type",
        "number_of_units_selected",
        "object_list_unit_id",
        "player_source",
        "area_1_x",
        "area_1_y",
        "area_2_x",
        "area_2_y",
        "object_group",
        "object_type",
        "object_list_unit_id_2",
        "selected_object_id",
    ],
    26: [
        "effect_type",
        "number_of_units_selected",
        "object_list_unit_id",
        "player_source",
        "string_id",
        "area_1_x",
        "area_1_y",
        "area_2_x",
        "area_2_y",
        "message",
        "selected_object_id",
    ],
    32: [
        "effect_type",
        "quantity",
        "number_of_units_selected",
        "object_list_unit_id",
        "player_source",
        "area_1_x",
        "area_1_y",
        "area_2_x",
        "area_2_y",
        "object_group",
        "object_type",
        "operation",
        "selected_object_id",
    ],
    33: [
        "effect_type",
        "quantity",
        "number_of_units_selected",
        "object_list_unit_id",
        "player_source",
        "area_1_x",
        "area_1_y",
        "area_2_x",
        "area_2_y",
        "object_group",
        "object_type",
        "selected_object_id",
    ],
    36: [
        "effect_type",
        "number_of_units_selected",
        "object_list_unit_id",
        "player_source",
        "area_1_x",
        "area_1_y",
        "area_2_x",
        "area_2_y",
        "object_group",
        "object_type",
        "attack_stance",
        "selected_object_id",
    ],
    18: [
        "effect_type",
        "number_of_units_selected",
        "object_list_unit_id",
        "player_source",
        "player_target",
        "area_1_x",
        "area_1_y",
        "area_2_x",
        "area_2_y",
        "object_group",
        "object_type",
        "flash_object",
        "selected_object_id",
    ],
    45: [
        "effect_type",
        "player_source",
        "string_id",
        "message",
    ],
    47: [
        "effect_type",
        "player_source",
        "technology",
        "object_list_unit_id_2",
        "button_location",
    ],
    46: [
        "effect_type",
        "object_list_unit_id",
        "player_source",
        "object_list_unit_id_2",
        "button_location",
    ],
    56: [
        "effect_type",
        "quantity",
        "operation",
        "from_variable",
        "message",
    ],
    16: [
        "effect_type",
        "player_source",
        "location_x",
        "location_y",
        "scroll",
    ],
    21: [
        "effect_type",
        "instruction_panel_position",
    ],
    57: [
        "effect_type",
        "variable_or_timer",
    ],
    49: [
        "effect_type",
        "number_of_units_selected",
        "object_list_unit_id",
        "player_source",
        "area_1_x",
        "area_1_y",
        "area_2_x",
        "area_2_y",
        "object_list_unit_id_2",
        "selected_object_id",
    ],
    11: [
        "effect_type",
        "object_list_unit_id",
        "player_source",
        "location_x",
        "location_y",
        "item_id",
        "facet",
    ],
    24: [
        "effect_type",
        "quantity",
        "number_of_units_selected",
        "object_list_unit_id",
        "player_source",
        "area_1_x",
        "area_1_y",
        "area_2_x",
        "area_2_y",
        "object_group",
        "object_type",
        "selected_object_id",
    ],
    9: [
        "effect_type",
        "trigger_id",
    ],
    13: [
        "effect_type",
        "player_source",
        "enabled_or_victory",
    ],
    20: [
        "effect_type",
        "object_list_unit_id",
        "player_source",
        "string_id",
        "display_time",
        "instruction_panel_position",
        "play_sound",
        "message",
        "sound_name",
    ],
    37: [
        "effect_type",
        "string_id",
        "display_time",
        "time_unit",
        "variable_or_timer",
        "message",
    ],
    38: [
        "effect_type",
        "object_list_unit_id",
        "player_source",
        "enabled_or_victory",
        "item_id",
    ],
    39: [
        "effect_type",
        "player_source",
        "technology",
        "enabled_or_victory",
        "item_id",
    ],
    22: [
        "effect_type",
        "number_of_units_selected",
        "object_list_unit_id",
        "player_source",
        "area_1_x",
        "area_1_y",
        "area_2_x",
        "area_2_y",
        "object_group",
        "object_type",
        "selected_object_id",
    ],
    34: [
        "effect_type",
        "quantity",
        "number_of_units_selected",
        "object_list_unit_id",
        "player_source",
        "area_1_x",
        "area_1_y",
        "area_2_x",
        "area_2_y",
        "object_group",
        "object_type",
        "selected_object_id",
    ],
    14: [
        "effect_type",
        "number_of_units_selected",
        "object_list_unit_id",
        "player_source",
        "area_1_x",
        "area_1_y",
        "area_2_x",
        "area_2_y",
        "object_group",
        "object_type",
        "selected_object_id",
    ],
    7: [
        "effect_type",
        "number_of_units_selected",
        "selected_object_id",
    ],
    51: [
        "effect_type",
        "quantity",
        "object_list_unit_id",
        "player_source",
        "item_id",
        "operation",
        "object_attributes",
    ],
    52: [
        "effect_type",
        "quantity",
        "tribute_list",
        "player_source",
        "item_id",
        "operation",
    ],
    53: [
        "effect_type",
        "tribute_list",
        "player_source",
        "item_id",
        "operation",
        "from_variable",
    ],
    19: [
        "effect_type",
        "number_of_units_selected",
        "object_list_unit_id",
        "player_source",
        "location_x",
        "location_y",
        "area_1_x",
        "area_1_y",
        "area_2_x",
        "area_2_y",
        "object_group",
        "object_type",
        "selected_object_id",
    ],
    25: [
        "effect_type",
        "object_list_unit_id",
        "player_source",
        "location_x",
        "location_y",
    ],
    4: [
        "effect_type",
        "player_source",
        "location_x",
        "location_y",
        "sound_name",
    ],
    15: [
        "effect_type",
        "number_of_units_selected",
        "object_list_unit_id",
        "player_source",
        "area_1_x",
        "area_1_y",
        "area_2_x",
        "area_2_y",
        "object_group",
        "object_type",
        "selected_object_id",
    ],
    43: [
        "effect_type",
        "number_of_units_selected",
        "object_list_unit_id",
        "player_source",
        "player_target",
        "area_1_x",
        "area_1_y",
        "area_2_x",
        "area_2_y",
        "object_group",
        "object_type",
        "object_list_unit_id_2",
        "selected_object_id",
    ],
    2: [
        "effect_type",
        "player_source",
        "technology",
        "force_research_technology",
    ],
    3: [
        "effect_type",
        "player_source",
        "string_id",
        "message",
        "sound_name",
    ],
    41: [
        "effect_type",
        "player_source",
        "player_target",
        "visibility_state",
    ],
    29: [
        "effect_type",
        "number_of_units_selected",
        "object_list_unit_id",
        "player_source",
        "area_1_x",
        "area_1_y",
        "area_2_x",
        "area_2_y",
        "object_group",
        "object_type",
        "selected_object_id",
    ],
    12: [
        "effect_type",
        "number_of_units_selected",
        "object_list_unit_id",
        "player_source",
        "location_x",
        "location_y",
        "area_1_x",
        "area_1_y",
        "area_2_x",
        "area_2_y",
        "object_group",
        "object_type",
        "selected_object_id",
    ],
    35: [
        "effect_type",
        "number_of_units_selected",
        "object_list_unit_id",
        "player_source",
        "location_x",
        "location_y",
        "area_1_x",
        "area_1_y",
        "area_2_x",
        "area_2_y",
        "object_group",
        "object_type",
        "selected_object_id",
    ],
    5: [
        "effect_type",
        "quantity",
        "tribute_list",
        "player_source",
        "player_target",
    ],
    17: [
        "effect_type",
        "number_of_units_selected",
        "object_list_unit_id",
        "player_source",
        "location_x",
        "location_y",
        "area_1_x",
        "area_1_y",
        "area_2_x",
        "area_2_y",
        "object_group",
        "object_type",
        "selected_object_id",
    ],
    6: [
        "effect_type",
        "number_of_units_selected",
        "selected_object_id",
    ],
    23: [
        "effect_type",
    ]
}

# for effect_id in range(0, 58):
#     try:
#         params = parameters[effect_id]
#         print("\"\"\"Attributes for the **xxxxx** effect are: \\n", end="")
#         for parameter in params:
#             if parameter is not "effect_type":
#                 print("\n-", parameter, end="")
#     except KeyError:
#         continue
#     print("\"\"\"")



from AoE2ScenarioParser.helper.retriever import find_retriever
from AoE2ScenarioParser.objects.aoe2_object import AoE2Object


class UnitObject(AoE2Object):
    def __init__(self,
                 x,
                 y,
                 z,
                 id_on_map,
                 unit_id,
                 status,
                 rotation,
                 animation_frame,
                 garrisoned_in_id
                 ):

        super().__init__(locals())

    @staticmethod
    def parse_object(parsed_data, **kwargs):  # Expected {unit=unitStruct}
        unit = kwargs['unit']

        return UnitObject(
            x=find_retriever(unit.retrievers, "X position").data,
            y=find_retriever(unit.retrievers, "Y position").data,
            z=find_retriever(unit.retrievers, "Z position").data,
            id_on_map=find_retriever(unit.retrievers, "ID").data,
            unit_id=find_retriever(unit.retrievers, "Unit 'constant'").data,
            status=find_retriever(unit.retrievers, "Status").data,
            rotation=find_retriever(unit.retrievers, "Rotation, in radians").data,
            animation_frame=find_retriever(unit.retrievers, "Initial animation frame").data,
            garrisoned_in_id=find_retriever(unit.retrievers, "Garrisoned in: ID").data,
        )

    @staticmethod
    def reconstruct_object(parsed_data, objects, **kwargs):
        pass



from AoE2ScenarioParser.datasets.conditions import condition_identifier_conversion
from AoE2ScenarioParser.helper import helper
from AoE2ScenarioParser.helper import parser
from AoE2ScenarioParser.helper.retriever import find_retriever
from AoE2ScenarioParser.objects.aoe2_object import AoE2Object
from AoE2ScenarioParser.objects.trigger_obj import TriggerObject


class TriggersObject(AoE2Object):
    def __init__(self,
                 trigger_data,
                 trigger_display_order
                 ):

        super().__init__(locals())

    def add_trigger(self, name):
        new_trigger = TriggerObject(name=name)
        self.data_dict['trigger_data'].append(new_trigger)
        helper.update_order_array(self.data_dict['trigger_display_order'], len(self.data_dict['trigger_data']))
        return new_trigger

    @staticmethod
    def parse_object(parsed_data, **kwargs):  # Expected {}
        display_order = parser.listify(
            find_retriever(parsed_data['TriggerPiece'].retrievers, "Trigger display order array").data)
        trigger_data = parser.listify(
            find_retriever(parsed_data['TriggerPiece'].retrievers, "Trigger data").data)

        triggers = []
        for trigger in trigger_data:
            triggers.append(TriggerObject.parse_object(parsed_data, trigger=trigger))

        return TriggersObject(
            trigger_data=triggers,
            trigger_display_order=display_order
        )

    @staticmethod
    def reconstruct_object(parsed_data, objects, **kwargs):  # Expected {}
        number_of_triggers_retriever = find_retriever(parsed_data['TriggerPiece'].retrievers, "Number of triggers")
        trigger_data_retriever = find_retriever(parsed_data['TriggerPiece'].retrievers, "Trigger data")
        display_order_retriever = find_retriever(parsed_data['TriggerPiece'].retrievers, "Trigger display order array")

        trigger_data_retriever.data = []
        for trigger in objects["TriggersObject"].data_dict['trigger_data']:
            TriggerObject.reconstruct_object(parsed_data, objects, trigger=trigger)

        assert len(trigger_data_retriever.data) == len(objects["TriggersObject"].data_dict['trigger_data'])
        trigger_count = len(trigger_data_retriever.data)
        number_of_triggers_retriever.data = trigger_count
        # Currently not necessary due to the parser setting repeated equal to list length
        # trigger_data_retriever.datatype.repeat = trigger_count
        # display_order_retriever.datatype.repeat = trigger_count
        display_order_retriever.data = parser.listify(display_order_retriever.data)
        helper.update_order_array(display_order_retriever.data, trigger_count)

    def get_trigger_overview_as_string(self):
        return_string = "Trigger Overview:\n"

        triggers = parser.listify(self.data_dict['trigger_data'])
        display_order = parser.listify(self.data_dict['trigger_display_order'])

        longest_trigger_name = -1
        for trigger_id in display_order:
            trigger_name = helper.del_str_trail(triggers[trigger_id].data_dict['name'])
            longest_trigger_name = max(longest_trigger_name, len(trigger_name))

        longest_trigger_name += 3
        for display, trigger_id in enumerate(display_order):
            trigger = triggers[trigger_id]
            trigger_data = trigger.data_dict
            trigger_name = helper.del_str_trail(trigger_data['name'])

            buffer = longest_trigger_name - len(trigger_name)
            return_string += "\t" + trigger_name + (" " * buffer)
            return_string += " [Index: " + str(trigger_id) + ", Display: " + str(display) + "]"

            return_string += "\t(conditions: " + str(len(parser.listify(trigger_data['conditions']))) + ", "
            return_string += " effects: " + str(len(parser.listify(trigger_data['effects']))) + ")\n"

        return return_string

    def get_content_as_string(self):
        return_string = "Triggers:\n"

        triggers = parser.listify(self.data_dict['trigger_data'])
        display_order = parser.listify(self.data_dict['trigger_display_order'])

        if len(triggers) == 0:
            return_string += "No triggers"
            return return_string

        for trigger_id in display_order:
            return_string += self.get_trigger_as_string(trigger_id) + "\n"

        return return_string

    def get_trigger_as_string(self, trigger_id):
        trigger = parser.listify(self.data_dict['trigger_data'])[trigger_id]
        display = parser.listify(self.data_dict['trigger_display_order']).index(trigger_id)
        trigger_data = trigger.data_dict

        return_string = ""

        return_string += "\t'" + helper.del_str_trail(trigger_data['name']) + "'"
        return_string += " [Index: " + str(trigger_id) + ", Display: " + str(display) + "]" + ":\n"

        return_string += trigger.get_content_as_string()

        return return_string

    def get_trigger_as_string_by_display_index(self, display_index):
        trigger_id = self.data_dict['trigger_display_order'][display_index]
        return self.get_trigger_as_string(trigger_id)

    def get_trigger(self, trigger_id):
        return parser.listify(self.data_dict['trigger_data'])[trigger_id]

    def get_trigger_by_display_index(self, display_index):
        trigger_id = self.data_dict['trigger_display_order'][display_index]
        return parser.listify(self.data_dict['trigger_data'])[trigger_id]

    def delete_trigger(self, trigger_id):
        del self.data_dict['trigger_data'][trigger_id]
        helper.update_order_array(self.data_dict['trigger_display_order'], len(self.data_dict['trigger_data']))
        del self.data_dict['trigger_display_order'][self.data_dict['trigger_display_order'].index(trigger_id)]
        self.data_dict['trigger_display_order'] = \
            [x - 1 if x > trigger_id else 0 for x in self.data_dict['trigger_display_order']]

    def delete_trigger_by_display_index(self, display_index):
        trigger_id = self.data_dict['trigger_display_order'][display_index]
        del self.data_dict['trigger_data'][trigger_id]
        del self.data_dict['trigger_display_order'][display_index]
        self.data_dict['trigger_display_order'] = \
            [x - 1 if x > trigger_id else 0 for x in self.data_dict['trigger_display_order']]



from AoE2ScenarioParser.objects.aoe2_object import AoE2Object


class FileHeaderObject(AoE2Object):
    def __init__(self,
                 version,
                 timestamp,
                 instructions,
                 player_count,
                 # Unknown data 36 bytes
                 creator_name,
                 # Unknown data 4 bytes
                 ):

        super().__init__(locals())

# For when turning it back into bytes:
# import time
# timestamp = int(time.time()) <-- for seconds since epoch



from AoE2ScenarioParser.objects.aoe2_object import AoE2Object


class OptionsObject(AoE2Object):
    def __init__(self,
                 disables,
                 all_techs
                 ):

        super().__init__(locals())



from AoE2ScenarioParser.objects.aoe2_object import AoE2Object
from AoE2ScenarioParser.helper.retriever import find_retriever
from AoE2ScenarioParser.objects.unit_obj import UnitObject


class UnitsObject(AoE2Object):
    def __init__(self,
                 units
                 ):

        super().__init__(locals())

    @staticmethod
    def parse_object(parsed_data, **kwargs):
        object_piece = parsed_data['UnitsPiece']
        units_per_player = find_retriever(object_piece.retrievers, "Player Units").data

        player_units = []

        for player_id in range(0, 9):  # 0 Gaia & 1-8 Players:
            player_units.append([])
            units = find_retriever(units_per_player[player_id].retrievers, "Units").data

            for unit in units:
                player_units[player_id].append(
                    UnitObject.parse_object(parsed_data, unit=unit)
                )

        return UnitsObject(
            units=player_units
        )

    @staticmethod
    def reconstruct_object(parsed_data, objects, **kwargs):
        pass



from AoE2ScenarioParser.helper.datatype import DataType
from AoE2ScenarioParser.helper.retriever import Retriever
import AoE2ScenarioParser.pieces.structs.aoe2_struct as structs


class ConditionStruct(structs.Struct):
    def __init__(self, parser_obj=None, data=None):
        retrievers = [
            Retriever("Condition type", DataType("s32")),
            Retriever("Check, (21)", DataType("s32")),  # Was always: 0x10
            Retriever("Amount (Quantity)", DataType("s32")),  # Also technology state, also Difficulty
            Retriever("Resource Type/Tribute list", DataType("s32")),
            Retriever("Unit object", DataType("s32")),
            Retriever("Next object", DataType("s32")),  # Old: Unit ID
            Retriever("Object list", DataType("s32")),
            Retriever("Player", DataType("s32")),
            Retriever("Technology", DataType("s32")),
            Retriever("Timer", DataType("s32")),
            Retriever("Unknown", DataType("s32")),
            Retriever("Area 1 X", DataType("s32")),
            Retriever("Area 1 Y", DataType("s32")),
            Retriever("Area 2 X", DataType("s32")),
            Retriever("Area 2 Y", DataType("s32")),
            Retriever("Object Group", DataType("s32")),
            Retriever("Object Type", DataType("s32")),  # -1: None, 1: Other, 2: Building, 3: Civilian, 4: Military
            Retriever("AI Signal", DataType("s32")),
            Retriever("Inverted", DataType("s32")),
            Retriever("Unknown (3)", DataType("s32")),
            Retriever("Variable", DataType("s32")),  # Number == VariableX
            Retriever("Comparison", DataType("s32")),  # 0: ==, 1: <, 2: >, 3: <=, 4 >=
            Retriever("Target player", DataType("s32")),
        ]

        super().__init__("Condition", retrievers, parser_obj, data)



from AoE2ScenarioParser.objects.aoe2_object import AoE2Object


class DataHeaderObject(AoE2Object):
    def __init__(self,
                 version,
                 filename,
                 ):

        super().__init__(locals())

    @staticmethod
    def parse_object(parsed_data, **kwargs):
        pass

    @staticmethod
    def reconstruct_object(parsed_data, objects, **kwargs):
        pass



import zlib
import collections

import AoE2ScenarioParser.helper.generator as generator
import AoE2ScenarioParser.helper.parser as parser
from AoE2ScenarioParser.objects.aoe2_object_manager import AoE2ObjectManager
from AoE2ScenarioParser.helper.datatype import DataType
from AoE2ScenarioParser.helper.helper import create_readable_hex_string
from AoE2ScenarioParser.helper.retriever import Retriever, find_retriever
from AoE2ScenarioParser.pieces.background_image import BackgroundImagePiece
from AoE2ScenarioParser.pieces.cinematics import CinematicsPiece
from AoE2ScenarioParser.pieces.data_header import DataHeaderPiece
from AoE2ScenarioParser.pieces.diplomacy import DiplomacyPiece
from AoE2ScenarioParser.pieces.options import OptionsPiece
from AoE2ScenarioParser.pieces.file_header import FileHeaderPiece
from AoE2ScenarioParser.pieces.global_victory import GlobalVictoryPiece
from AoE2ScenarioParser.pieces.map import MapPiece
from AoE2ScenarioParser.pieces.messages import MessagesPiece
from AoE2ScenarioParser.pieces.player_data_two import PlayerDataTwoPiece
from AoE2ScenarioParser.pieces.triggers import TriggerPiece
from AoE2ScenarioParser.pieces.units import UnitsPiece
from AoE2ScenarioParser.datasets import effects as effect_dataset
from AoE2ScenarioParser.datasets import conditions as condition_dataset


class AoE2Scenario:
    def __init__(self, filename):
        print("\nLoading file: '" + filename + "'...")

        scenario = open(filename, "rb")
        self.file = scenario.read()
        scenario.seek(0)  # Reset file cursor to 0
        self.file_header = scenario.read(self._compute_header_length())
        self.file_data = zlib.decompress(scenario.read(), -zlib.MAX_WBITS)
        scenario.close()

        print("File loaded.")

        self.parser = parser.Parser()
        self._read_file()

    def _read_file(self):
        print("\nFile reading started...")
        self.parsed_header = collections.OrderedDict()
        self.parsed_data = collections.OrderedDict()
        header_generator = self._create_header_generator(1)
        data_generator = self._create_data_generator(1)

        structures = [_header_structure, _file_structure]
        structure_generators = [header_generator, data_generator]
        structure_parsed = [self.parsed_header, self.parsed_data]

        for index, structure in enumerate(structures):
            for piece_object in structure:
                piece = piece_object(self.parser)
                print("Reading", piece.piece_type + "...")
                piece.set_data_from_generator(structure_generators[index])
                structure_parsed[index][type(piece).__name__] = piece
                print("Reading", piece.piece_type, "finished successfully.")

        suffix = b''
        try:
            while True:
                suffix += self.parser.retrieve_value(data_generator, Retriever("suffix data", DataType("1")))
        except StopIteration as e:
            # print(e)
            pass
        finally:
            # print("Suffix done", len(suffix))
            self.suffix = suffix

        print("File reading finished successfully.")

        self.object_manager = AoE2ObjectManager(self.parsed_header, self.parsed_data)

    def write_to_file(self, filename):
        self._write_from_structure(filename)

    def _write_from_structure(self, filename, write_in_bytes=True):
        self.object_manager.reconstruct()
        print("\nFile writing from structure started...")
        byte_header = b''
        byte_data = b''

        for key in self.parsed_header:
            for retriever in self.parsed_header[key].retrievers:
                byte_header += parser.retriever_to_bytes(retriever)

        for key in self.parsed_data:
            for retriever in self.parsed_data[key].retrievers:
                try:
                    byte_data += parser.retriever_to_bytes(retriever)
                except AttributeError:
                    print(key, retriever)

        # https://stackoverflow.com/questions/3122145/zlib-error-error-3-while-decompressing-incorrect-header-check/22310760#22310760
        deflate_obj = zlib.compressobj(9, zlib.DEFLATED, -zlib.MAX_WBITS)
        compressed = deflate_obj.compress(byte_data + self.suffix) + deflate_obj.flush()

        file = open(filename, "wb" if write_in_bytes else "w")
        file.write(byte_header if write_in_bytes else create_readable_hex_string(byte_header.hex()))
        file.write(compressed if write_in_bytes else create_readable_hex_string(compressed.hex()))
        file.close()
        print("File writing finished successfully.")

    def write_file(self, datatype, write_in_bytes=True):
        print("File writing from source started with attributes " + datatype + "...")
        file = open("./../results/generated_map_" + datatype + ".aoe2scenario", "wb" if write_in_bytes else "w")
        for t in datatype:
            if t == "f":
                file.write(self.file if write_in_bytes else create_readable_hex_string(self.file.hex()))
            elif t == "h":
                file.write(self.file_header if write_in_bytes else create_readable_hex_string(self.file_header.hex()))
            elif t == "d":
                file.write(self.file_data if write_in_bytes else create_readable_hex_string(self.file_data.hex()))
        file.close()
        print("File writing finished successfully.")

    def _create_header_generator(self, chunk_size):
        return generator.create_generator(self.file_header, chunk_size)

    def _create_data_generator(self, chunk_size):
        return generator.create_generator(self.file_data, chunk_size)

    def _create_file_generator(self, chunk_size):
        return generator.create_generator(self.file, chunk_size)

    def _compute_header_length(self):
        return parser.calculate_length(
            self._create_file_generator(1),
            FileHeaderPiece(parser.Parser()).retrievers
        )

    def _log_all_data_keys(self):
        """ Used for debugging. """
        print("FileHeader:")
        print(self.parsed_header.keys())
        print("DataHeader:")
        print("odict_keys([")
        for x in self.parsed_data.keys():
            print("\t'" + x + "'")
        print("])")

    def _log_effect_dataset(self):
        """ Used for debugging - Only reads One Trigger. """
        trigger_data = find_retriever(self.parsed_data['TriggerPiece'].retrievers, "Trigger data").data
        effects = parser.listify(find_retriever(trigger_data.retrievers, "Effect data").data)

        for effect in effects:
            for retriever in effect.retrievers:
                if retriever.data != -1 and \
                        retriever.data != [] and \
                        retriever.data != "" and \
                        retriever.data != " " and \
                        retriever.name != "Check, (46)":
                    if retriever.name == "Effect type":
                        print("],\n" + str(retriever.data) + ": [")
                    # print(retriever)
                    print("\t\"" + str(effect_dataset.attribute_naming_conversion[retriever.name]) + "\",")
        print("]\n")

    def _log_condition_dataset(self):
        """ Used for debugging - Only reads One Trigger. """
        trigger_data = find_retriever(self.parsed_data['TriggerPiece'].retrievers, "Trigger data").data
        conditions = parser.listify(find_retriever(trigger_data.retrievers, "Condition data").data)

        for condition in conditions:
            for retriever in condition.retrievers:
                if retriever.data != -1 and \
                        retriever.data != [] and \
                        retriever.data != "" and \
                        retriever.data != " " and \
                        retriever.name != "Check, (21)":
                    # print(retriever)
                    if retriever.name == "Condition type":
                        print("],\n" + str(retriever.data) + ": [")
                    print("\t\"" + str(condition_dataset.attribute_naming_conversion[retriever.name]) + "\",")
        print("]\n")


_header_structure = [
    FileHeaderPiece
]
_file_structure = [
    DataHeaderPiece,
    MessagesPiece,
    CinematicsPiece,
    BackgroundImagePiece,
    PlayerDataTwoPiece,
    GlobalVictoryPiece,
    DiplomacyPiece,
    OptionsPiece,
    MapPiece,
    UnitsPiece,
    TriggerPiece,
]



def update_order_array(order_array, supposed_length):
    for i in range(0, supposed_length):
        if i not in order_array:
            order_array.append(i)


def create_readable_hex_string(string):
    return insert_char(insert_char(string, " ", 2), "\n", 48)


# Credits: gurney alex @ https://stackoverflow.com/a/2657733/7230293
def insert_char(string, char, every=64):
    return char.join(string[i:i + every] for i in range(0, len(string), every))


def del_str_trail(string):
    return string.replace('\x00', "")




none = 0
"""Attributes for the **none**  condition are:\n
... It's called none... What parameters do **you** think it has?! ❤ """
bring_object_to_area = 1
"""Attributes for the **bring_object_to_area** condition are:\n
- unit_object  
- area_1_x
- area_1_y
- area_2_x
- area_2_y
- inverted"""
bring_object_to_object = 2
"""Attributes for the **bring_object_to_object** condition are:\n
- unit_object
- next_object
- inverted"""
own_objects = 3
"""Attributes for the **own_objects** condition are:\n
- amount_or_quantity
- object_list
- player
- object_group
- object_type"""
owh_fewer_objects = 4
"""Attributes for the **owh_fewer_objects** condition are:\n
- amount_or_quantity
- object_list
- player
- area_1_x
- area_1_y
- area_2_x
- area_2_y
- object_group
- object_type"""
object_in_area = 5
"""Attributes for the **object_in_area** condition are:\n
- amount_or_quantity
- object_list
- player
- area_1_x
- area_1_y
- area_2_x
- area_2_y
- object_group
- object_type
- inverted"""
destroy_object = 6
"""Attributes for the **destroy_object** condition are:\n
- unit_object
- inverted"""
capture_object = 7
"""Attributes for the **capture_object** condition are:\n
- unit_object
- player
- inverted"""
accumulate_attribute = 8
"""Attributes for the **accumulate_attribute** condition are:\n
- amount_or_quantity
- resource_type_or_tribute_list
- player
- inverted"""
research_technology = 9
"""Attributes for the **research_technology** condition are:\n
- player
- technology
- inverted"""
timer = 10
"""Attributes for the **timer** condition are:\n
- timer
- inverted"""
object_selected = 11
"""Attributes for the **object_selected** condition are:\n
- unit_object
- inverted"""
ai_signal = 12
"""Attributes for the **ai_signal** condition are:\n
- ai_signal
- inverted"""
player_defeated = 13
"""Attributes for the **player_defeated** condition are:\n
- player
- inverted"""
object_has_target = 14
"""Attributes for the **object_has_target** condition are:\n
- unit_object
- next_object
- object_list
- object_group
- object_type
- inverted"""
object_visible = 15
"""Attributes for the **object_visible** condition are:\n
- unit_object"""
object_not_visible = 16
"""Attributes for the **object_not_visible** condition are:\n
- unit_object"""
researching_tech = 17
"""Attributes for the **researching_tech** condition are:\n
- player
- technology
- inverted"""
units_garrisoned = 18
"""Attributes for the **units_garrisoned** condition are:\n
- amount_or_quantity
- unit_object
- inverted"""
chance = 20
"""Attributes for the **chance** condition are:\n
- amount_or_quantity"""
technology_state = 21
"""Attributes for the **technology_state** condition are:\n
- amount_or_quantity
- player
- technology
- inverted"""
variable_value = 22
"""Attributes for the **variable_value** condition are:\n
- amount_or_quantity
- inverted
- variable
- comparison"""
object_hp = 23
"""Attributes for the **object_hp** condition are:\n
- amount_or_quantity
- unit_object
- inverted
- comparison"""
diplomacy_state = 24
"""Attributes for the **diplomacy_state** condition are:\n
- amount_or_quantity
- player
- inverted
- target_player"""

attribute_naming_conversion = {
    "condition_type": "Condition type",
    "amount_or_quantity": "Amount (Quantity)",
    "resource_type_or_tribute_list": "Resource Type/Tribute list",
    "unit_object": "Unit object",
    "next_object": "Next object",
    "object_list": "Object list",
    "player": "Player",
    "technology": "Technology",
    "timer": "Timer",
    "area_1_x": "Area 1 X",
    "area_1_y": "Area 1 Y",
    "area_2_x": "Area 2 X",
    "area_2_y": "Area 2 Y",
    "object_group": "Object Group",
    "object_type": "Object Type",
    "ai_signal": "AI Signal",
    "inverted": "Inverted",
    "variable": "Variable",
    "comparison": "Comparison",
    "target_player": "Target player",
}

condition_identifier_conversion = {
    0: "none",
    1: "bring_object_to_area",
    2: "bring_object_to_object",
    3: "own_objects",
    4: "own_fewer_objects",
    5: "objects_in_area",
    6: "destroy_object",
    7: "capture_object",
    8: "accumulate_attribute",
    9: "research_technology",
    10: "timer",
    11: "object_selected",
    12: "ai_signal",
    13: "player_defeated",
    14: "object_has_target",
    15: "object_visible",
    16: "object_not_visible",
    17: "researching_tech",
    18: "units_garrisoned",
    20: "chance",
    21: "technology_state",
    22: "variable_value",
    23: "object_hp",
    24: "diplomacy_state",
}

# Credits: Eti JS @ https://stackoverflow.com/a/58746861/7230293
attribute_naming_conversion.update(dict(map(reversed, attribute_naming_conversion.items())))
condition_identifier_conversion.update(dict(map(reversed, condition_identifier_conversion.items())))

empty_attributes = {
    "condition_type": -1,
    "amount_or_quantity": -1,
    "resource_type_or_tribute_list": -1,
    "unit_object": -1,
    "next_object": -1,
    "object_list": -1,
    "player": -1,
    "technology": -1,
    "timer": -1,
    "area_1_x": -1,
    "area_1_y": -1,
    "area_2_x": -1,
    "area_2_y": -1,
    "object_group": -1,
    "object_type": -1,
    "ai_signal": -1,
    "inverted": -1,
    "variable": -1,
    "comparison": -1,
    "target_player": -1,
}

attributes = {
    0: [
        "condition_type",
    ],
    8: [
        "condition_type",
        "amount_or_quantity",
        "resource_type_or_tribute_list",
        "player",
        "inverted",
    ],
    12: [
        "condition_type",
        "ai_signal",
        "inverted",
    ],
    1: [
        "condition_type",
        "unit_object",
        "area_1_x",
        "area_1_y",
        "area_2_x",
        "area_2_y",
        "inverted",
    ],
    2: [
        "condition_type",
        "unit_object",
        "next_object",
        "inverted",
    ],
    7: [
        "condition_type",
        "unit_object",
        "player",
        "inverted",
    ],
    20: [
        "condition_type",
        "amount_or_quantity",
    ],
    6: [
        "condition_type",
        "unit_object",
        "inverted",
    ],
    24: [
        "condition_type",
        "amount_or_quantity",
        "player",
        "inverted",
        "target_player",
    ],
    14: [
        "condition_type",
        "unit_object",
        "next_object",
        "object_list",
        "object_group",
        "object_type",
        "inverted",
    ],
    23: [
        "condition_type",
        "amount_or_quantity",
        "unit_object",
        "inverted",
        "comparison",
    ],
    16: [
        "condition_type",
        "unit_object",
    ],
    11: [
        "condition_type",
        "unit_object",
        "inverted",
    ],
    15: [
        "condition_type",
        "unit_object",
    ],
    5: [
        "condition_type",
        "amount_or_quantity",
        "object_list",
        "player",
        "area_1_x",
        "area_1_y",
        "area_2_x",
        "area_2_y",
        "object_group",
        "object_type",
        "inverted",
    ],
    4: [
        "condition_type",
        "amount_or_quantity",
        "object_list",
        "player",
        "area_1_x",
        "area_1_y",
        "area_2_x",
        "area_2_y",
        "object_group",
        "object_type",
    ],
    3: [
        "condition_type",
        "amount_or_quantity",
        "object_list",
        "player",
        "object_group",
        "object_type",
    ],
    13: [
        "condition_type",
        "player",
        "inverted",
    ],
    9: [
        "condition_type",
        "player",
        "technology",
        "inverted",
    ],
    17: [
        "condition_type",
        "player",
        "technology",
        "inverted",
    ],
    21: [
        "condition_type",
        "amount_or_quantity",
        "player",
        "technology",
        "inverted",
    ],
    10: [
        "condition_type",
        "timer",
        "inverted",
    ],
    18: [
        "condition_type",
        "amount_or_quantity",
        "unit_object",
        "inverted",
    ],
    22: [
        "condition_type",
        "amount_or_quantity",
        "inverted",
        "variable",
        "comparison",
    ]
}

# for condition_id in range(0, 25):
#     try:
#         params = parameters[condition_id]
#         print("\"\"\"Attributes for the **xxxxx** condition are:\\n", end="")
#         for parameter in params:
#             if parameter is not "condition_type":
#                 print("\n-", parameter, end="")
#     except KeyError:
#         continue
#     print("\"\"\"")



import AoE2ScenarioParser.pieces.aoe2_piece as scenario_piece
from AoE2ScenarioParser.helper.retriever import Retriever
from AoE2ScenarioParser.helper.datatype import DataType
from AoE2ScenarioParser.pieces.structs.trigger import TriggerStruct


class TriggerPiece(scenario_piece.ScenarioPiece):
    def __init__(self, parser_obj=None, data=None):
        retrievers = [
            Retriever("Trigger Version", DataType("f64")),
            Retriever("Trigger instructions start", DataType("s8")),
            Retriever("Number of triggers", DataType("s32"), save_as="number_of_triggers"),
            Retriever("Trigger data", DataType(TriggerStruct), set_repeat="{number_of_triggers}"),
            Retriever("Trigger display order array", DataType("u32"), set_repeat="{number_of_triggers}"),
        ]

        super().__init__("Triggers", retrievers, parser_obj)



from AoE2ScenarioParser.helper.datatype import DataType
from AoE2ScenarioParser.helper.retriever import Retriever
import AoE2ScenarioParser.pieces.structs.aoe2_struct as structs


class EffectStruct(structs.Struct):
    def __init__(self, parser_obj=None, data=None):
        retrievers = [
            Retriever("Effect type", DataType("s32"), save_as="effect_type"),                               # CONFIRMED
            Retriever("Check, (46)", DataType("s32")),  # always 0x17, now 0x2e (46)?                       # CONFIRMED
            Retriever("AI script goal", DataType("s32")),                                                   # CONFIRMED
            Retriever("AA Quantity", DataType("u8"),
                      set_repeat="1 if {effect_type} == 31 or {effect_type} == 28 else 0"),                 # CONFIRMED
            Retriever("AA Armor/Attack Type", DataType("s24"),
                      set_repeat="1 if {effect_type} == 31 or {effect_type} == 28 else 0"),                 # CONFIRMED
            Retriever("Quantity", DataType("s32"),
                      set_repeat="1 if {effect_type} != 31 and {effect_type} != 28 else 0"),                # CONFIRMED
            Retriever("Tribute List", DataType("s32")),                                                     # CONFIRMED
            Retriever("Diplomacy", DataType("s32")),                                                        # CONFIRMED
            Retriever("Number of units selected", DataType("s32"), save_as="number_of_units_selected"),     # CONFIRMED
            Retriever("Unknown", DataType("s32")),                                                          # UNUSED
            Retriever("Object list unit ID", DataType("s32")),                                              # CONFIRMED
            Retriever("Player Source", DataType("s32")),                                                    # CONFIRMED
            Retriever("Player Target", DataType("s32")),                                                    # CONFIRMED
            Retriever("Technology", DataType("s32")),                                                       # CONFIRMED
            Retriever("String ID", DataType("s32")),                                                        # CONFIRMED
            Retriever("Unknown2", DataType("s32")),                                                         # UNUSED
            Retriever("Display Time (display instructions)", DataType("s32")),                              # CONFIRMED
            Retriever("Trigger ID (activate/deactivate)", DataType("s32")),                                 # CONFIRMED
            Retriever("Location X", DataType("s32")),                                                       # CONFIRMED
            Retriever("Location Y", DataType("s32")),                                                       # CONFIRMED
            Retriever("Area 1 X", DataType("s32")),                                                         # CONFIRMED
            Retriever("Area 1 Y", DataType("s32")),                                                         # CONFIRMED
            Retriever("Area 2 X", DataType("s32")),                                                         # CONFIRMED
            Retriever("Area 2 Y", DataType("s32")),                                                         # CONFIRMED
            Retriever("Object Group", DataType("s32")),                                                     # CONFIRMED
            Retriever("Object Type", DataType("s32")),                                                      # CONFIRMED
            Retriever("Instruction Panel Position", DataType("s32")),                                       # CONFIRMED
            Retriever("Attack Stance", DataType("s32")),                                                    # CONFIRMED
            Retriever("Time unit (second, minutes, years)", DataType("s32")),                               # CONFIRMED
            Retriever("Enabled/Victory", DataType("s32")),                                                  # CONFIRMED
            Retriever("Food", DataType("s32")),                                                             # CONFIRMED
            Retriever("Wood", DataType("s32")),                                                             # CONFIRMED
            Retriever("Stone", DataType("s32")),                                                            # CONFIRMED
            Retriever("Gold", DataType("s32")),                                                             # CONFIRMED
            Retriever("Item ID", DataType("s32")),                                                          # CONFIRMED
            Retriever("Flash Object", DataType("s32")),                                                     # CONFIRMED
            Retriever("Force Research Technology", DataType("s32")),                                        # CONFIRMED
            Retriever("Visibility State", DataType("s32")),                                                 # CONFIRMED
            Retriever("Scroll (Set view)", DataType("s32")),                                                # CONFIRMED
            Retriever("Operation", DataType("s32")),                                                        # CONFIRMED
            Retriever("Object list unit ID 2", DataType("s32")),                                            # CONFIRMED
            Retriever("Button Location", DataType("s32")),                                                  # CONFIRMED
            Retriever("AI signal Value", DataType("s32")),                                                  # CONFIRMED
            Retriever("Unknown3", DataType("s32")),                                                         # UNUSED
            Retriever("Object attributes", DataType("s32")),                                                # CONFIRMED
            Retriever("From Variable", DataType("s32")),                                                    # CONFIRMED
            Retriever("Variable/Timer", DataType("s32")),                                                   # CONFIRMED
            Retriever("Facet", DataType("s32")),                                                            # CONFIRMED
            Retriever("Unknown4", DataType("s32")),                                                         # UNUSED
            Retriever("Play Sound", DataType("s32")),                                                       # CONFIRMED
            Retriever("Message", DataType("str32")),                                                        # CONFIRMED
            Retriever("Sound (event) name", DataType("str32")),                                             # CONFIRMED
            Retriever("Selected Object(s) ID", DataType("s32"), set_repeat="{number_of_units_selected}"),   # CONFIRMED
        ]

        super().__init__("Effect", retrievers, parser_obj, data)



import abc


class AoE2Object:
    def __init__(self, data_dict):
        del data_dict['self']
        del data_dict['__class__']

        self.data_dict = data_dict

        # Block terrain, units and effects from logging (respectively)
        # if 'terrain_id' not in self.__str__() and \
        #         'id_on_map' not in self.__str__() and \
        #         'Effect type' not in self.__str__():
        #     print(data_dict)

    @staticmethod
    @abc.abstractmethod
    def parse_object(parsed_data, **kwargs):
        pass

    @staticmethod
    @abc.abstractmethod
    def reconstruct_object(parsed_data, objects, **kwargs):
        pass

    def __str__(self):
        return str(self.data_dict)



import copy

from AoE2ScenarioParser.datasets import conditions
from AoE2ScenarioParser.helper.retriever import find_retriever
from AoE2ScenarioParser.objects.aoe2_object import AoE2Object
from AoE2ScenarioParser.pieces.structs.condition import ConditionStruct


class ConditionObject(AoE2Object):
    def __init__(self,
                 condition_type,
                 amount_or_quantity=-1,
                 resource_type_or_tribute_list=-1,
                 unit_object=-1,
                 next_object=-1,
                 object_list=-1,
                 player=1,
                 technology=-1,
                 timer=-1,
                 area_1_x=-1,
                 area_1_y=-1,
                 area_2_x=-1,
                 area_2_y=-1,
                 object_group=-1,
                 object_type=-1,
                 ai_signal=-1,
                 inverted=0,
                 variable=-1,
                 comparison=-1,
                 target_player=-1
                 ):

        super().__init__(locals())

    def get_content_as_string(self):
        return_string = ""

        for attribute in conditions.attributes[self.data_dict['condition_type']]:
            if attribute == "condition_type" or self.data_dict[attribute] == [] or self.data_dict[attribute] == "" or \
                    self.data_dict[attribute] == " " or self.data_dict[attribute] == -1:
                continue
            return_string += "\t\t\t\t" + attribute + ": " + str(self.data_dict[attribute]) + "\n"

        return return_string

    @staticmethod
    def parse_object(parsed_data, **kwargs):  # Expected {condition=conditionStruct}
        condition_struct = kwargs['condition']

        effect_type = find_retriever(condition_struct.retrievers, "Condition type").data
        parameters = conditions.attributes.get(effect_type)

        parameter_dict = copy.copy(conditions.empty_attributes)
        for param in parameters:
            parameter_dict[param] = find_retriever(condition_struct.retrievers,
                                                   conditions.attribute_naming_conversion[param]).data

        return ConditionObject(
            **parameter_dict
        )

    @staticmethod
    def reconstruct_object(parsed_data, objects, **kwargs):
        # Expected {condition=condition_obj, conditions=conditionsList}
        condition_obj = kwargs['condition']
        conditions = kwargs['conditions']

        data_list = list(condition_obj.data_dict.values())
        data_list.insert(1, 21)  # Check, (21)
        data_list.insert(10, -1)  # Unknown
        data_list.insert(19, -1)  # Unknown (3)
        conditions.append(ConditionStruct(data=data_list))

    def set_condition_type(self, val):
        self.data_dict['condition_type'] = val

    def set_amount_or_quantity(self, val):
        self.data_dict['amount_or_quantity'] = val

    def set_resource_type_or_tribute_list(self, val):
        self.data_dict['resource_type_or_tribute_list'] = val

    def set_unit_object(self, val):
        self.data_dict['unit_object'] = val

    def set_next_object(self, val):
        self.data_dict['next_object'] = val

    def set_object_list(self, val):
        self.data_dict['object_list'] = val

    def set_player(self, val):
        self.data_dict['player'] = val

    def set_technology(self, val):
        self.data_dict['technology'] = val

    def set_timer(self, val):
        self.data_dict['timer'] = val

    def set_area_1_x(self, val):
        self.data_dict['area_1_x'] = val

    def set_area_1_y(self, val):
        self.data_dict['area_1_y'] = val

    def set_area_2_x(self, val):
        self.data_dict['area_2_x'] = val

    def set_area_2_y(self, val):
        self.data_dict['area_2_y'] = val

    def set_object_group(self, val):
        self.data_dict['object_group'] = val

    def set_object_type(self, val):
        self.data_dict['object_type'] = val

    def set_ai_signal(self, val):
        self.data_dict['ai_signal'] = val

    def set_inverted(self, val):
        self.data_dict['inverted'] = val

    def set_variable(self, val):
        self.data_dict['variable'] = val

    def set_comparison(self, val):
        self.data_dict['comparison'] = val

    def set_target_player(self, val):
        self.data_dict['target_player'] = val



import AoE2ScenarioParser.pieces.aoe2_piece as scenario_piece
from AoE2ScenarioParser.helper.retriever import Retriever
from AoE2ScenarioParser.helper.datatype import DataType


class FileHeaderPiece(scenario_piece.ScenarioPiece):
    def __init__(self, parser_obj=None, data=None):
        retrievers = [
            Retriever("Version", DataType("c4")),
            Retriever("Header length", DataType("u32")),
            Retriever("Savable", DataType("s32")),
            Retriever("Timestamp of last save", DataType("u32")),
            Retriever("Scenario instructions", DataType("str32")),
            Retriever("Individual victories used", DataType("u32")),
            Retriever("Player count", DataType("u32")),
            Retriever("Data", DataType("36")),
            Retriever("Creator name", DataType("str32")),
            Retriever("Data", DataType("4")),
        ]

        super().__init__("File Header", retrievers, parser_obj)



from AoE2ScenarioParser.objects.aoe2_object import AoE2Object


class MapObject(AoE2Object):
    def __init__(self,
                 map_color_mood,
                 collide_and_correct,
                 map_width,
                 map_height,
                 terrain
                 ):

        super().__init__(locals())



from AoE2ScenarioParser.objects.aoe2_object import AoE2Object


class MessagesObject(AoE2Object):
    def __init__(self,
                 instructions,
                 hints,
                 victory,
                 loss,
                 history,
                 scouts,
                 ascii_instructions,
                 ascii_hints,
                 ascii_victory,
                 ascii_loss,
                 ascii_history,
                 ascii_scouts,
                 ):

        super().__init__(locals())



from AoE2ScenarioParser.objects.aoe2_object import AoE2Object


class PlayerObject(AoE2Object):
    def __init__(self,
                 player_number,
                 active,
                 human,
                 civilization,
                 gold,
                 wood,
                 food,
                 stone,
                 color,
                 starting_age,
                 pop_limit
                 ):

        super().__init__(locals())



from AoE2ScenarioParser.datasets.conditions import condition_identifier_conversion
from AoE2ScenarioParser.datasets.effects import effect_identifier_conversion
from AoE2ScenarioParser.helper import helper
from AoE2ScenarioParser.helper import parser
from AoE2ScenarioParser.helper.retriever import find_retriever
from AoE2ScenarioParser.objects.aoe2_object import AoE2Object
from AoE2ScenarioParser.objects.condition_obj import ConditionObject
from AoE2ScenarioParser.objects.effect_obj import EffectObject
from AoE2ScenarioParser.pieces.structs.condition import ConditionStruct
from AoE2ScenarioParser.pieces.structs.trigger import TriggerStruct


class TriggerObject(AoE2Object):
    def __init__(self,
                 name,
                 description="",
                 description_stid=0,
                 display_as_objective=0,
                 short_description="",
                 short_description_stid=0,
                 display_on_screen=0,
                 description_order=0,
                 enabled=1,
                 looping=0,
                 header=0,
                 mute_objectives=0,
                 conditions=None,
                 condition_order=None,
                 effects=None,
                 effect_order=None,
                 ):

        # Adding expected (by DE) ending character
        name += "\x00"

        if conditions is None:
            conditions = []
        if condition_order is None:
            condition_order = []
        if effects is None:
            effects = []
        if effect_order is None:
            effect_order = []

        super().__init__(locals())

    def get_content_as_string(self):
        return_string = ""
        data_tba = [
            ('enabled', self.data_dict['enabled'] != 0),
            ('looping', self.data_dict['looping'] != 0)
        ]

        if helper.del_str_trail(self.data_dict['description']) != "":
            data_tba.append(('description', "'" + helper.del_str_trail(self.data_dict['description']) + "'"))
        if self.data_dict['description_stid'] != 0:
            data_tba.append(('description_stid', self.data_dict['description_stid']))
        if helper.del_str_trail(self.data_dict['short_description']) != "":
            data_tba.append(('short_description', "'" + helper.del_str_trail(self.data_dict['short_description']) + "'"))
        if self.data_dict['short_description_stid'] != 0:
            data_tba.append(('short_description_stid', self.data_dict['short_description_stid']))
        if self.data_dict['display_as_objective'] != 0:
            data_tba.append(('display_as_objective', self.data_dict['display_as_objective'] != 0))
        if self.data_dict['display_on_screen'] != 0:
            data_tba.append(('display_on_screen', self.data_dict['display_on_screen'] != 0))
        if self.data_dict['description_order'] != 0:
            data_tba.append(('description_order', self.data_dict['description_order']))
        if self.data_dict['header'] != 0:
            data_tba.append(('header', self.data_dict['header'] != 0))
        if self.data_dict['mute_objectives'] != 0:
            data_tba.append(('mute_objectives', self.data_dict['mute_objectives'] != 0))

        for data in data_tba:
            return_string += "\t\t" + data[0] + ": " + str(data[1]) + "\n"

        if len(self.data_dict['condition_order']) > 0:
            return_string += "\t\tconditions:\n"
            for c_display_order, condition_id in enumerate(self.data_dict['condition_order']):
                condition = self.data_dict['conditions'][condition_id]

                return_string += "\t\t\t" + condition_identifier_conversion[condition.data_dict['condition_type']] + ":\n"
                return_string += condition.get_content_as_string()

        if len(self.data_dict['effect_order']) > 0:
            return_string += "\t\teffects:\n"
            for e_display_order, effect_id in enumerate(self.data_dict['effect_order']):
                effect = self.data_dict['effects'][effect_id]

                return_string += "\t\t\t" + effect_identifier_conversion[effect.data_dict['effect_type']] + ":\n"
                return_string += effect.get_content_as_string()

        return return_string

    def add_effect(self, effect_type):
        new_effect = EffectObject(effect_type)
        self.data_dict['effects'].append(new_effect)
        return new_effect

    def add_condition(self, condition_type):
        new_cond = ConditionObject(condition_type)
        self.data_dict['conditions'].append(new_cond)
        return new_cond

    @staticmethod
    def parse_object(parsed_data, **kwargs):  # Expected {trigger=triggerStruct}
        trigger = kwargs['trigger']

        effects = []
        effect_structs = parser.listify(find_retriever(trigger.retrievers, "Effect data").data)
        for effect_struct in effect_structs:
            effects.append(EffectObject.parse_object(parsed_data, effect=effect_struct))

        conditions = []
        condition_structs = parser.listify(find_retriever(trigger.retrievers, "Condition data").data)
        for condition_struct in condition_structs:
            conditions.append(ConditionObject.parse_object(parsed_data, condition=condition_struct))

        return TriggerObject(
            name=find_retriever(trigger.retrievers, "Trigger name").data,
            description=find_retriever(trigger.retrievers, "Trigger description").data,
            description_stid=find_retriever(trigger.retrievers, "Description string Table ID").data,
            display_as_objective=find_retriever(trigger.retrievers, "Act as objective").data,
            short_description=find_retriever(trigger.retrievers, "Short description").data,
            short_description_stid=find_retriever(trigger.retrievers, "Short description string Table ID").data,
            display_on_screen=find_retriever(trigger.retrievers, "Display on screen").data,
            description_order=find_retriever(trigger.retrievers, "Description order (in objectives)").data,
            enabled=find_retriever(trigger.retrievers, "Enabled").data,
            looping=find_retriever(trigger.retrievers, "Looping").data,
            header=find_retriever(trigger.retrievers, "Make header").data,
            mute_objectives=find_retriever(trigger.retrievers, "Mute objectives").data,
            conditions=conditions,
            condition_order=parser.listify(find_retriever(trigger.retrievers, "Condition display order array").data),
            effects=effects,
            effect_order=parser.listify(find_retriever(trigger.retrievers, "Effect display order array").data),
        )

    @staticmethod
    def reconstruct_object(parsed_data, objects, **kwargs):  # Expected {trigger=triggerStruct}
        trigger_data_retriever = find_retriever(parsed_data['TriggerPiece'].retrievers, "Trigger data")
        trigger = kwargs['trigger']

        effects = []
        for effect_obj in trigger.data_dict['effects']:
            EffectObject.reconstruct_object(parsed_data, objects, effect=effect_obj, effects=effects)

        helper.update_order_array(
            parser.listify(trigger.data_dict['effect_order']), len(trigger.data_dict['effects']))

        conditions = []
        for condition_obj in trigger.data_dict['conditions']:
            data_list = list(condition_obj.data_dict.values())
            data_list.insert(1, 21)  # Check, (21)
            data_list.insert(10, -1)  # Unknown
            data_list.insert(19, -1)  # Unknown (3)
            conditions.append(ConditionStruct(data=data_list))

        helper.update_order_array(
            parser.listify(trigger.data_dict['condition_order']), len(trigger.data_dict['conditions']))

        trigger_data_retriever.data.append(TriggerStruct(data=[
            trigger.data_dict['enabled'],
            trigger.data_dict['looping'],
            trigger.data_dict['description_stid'],
            trigger.data_dict['display_as_objective'],
            trigger.data_dict['description_order'],
            trigger.data_dict['header'],
            trigger.data_dict['short_description_stid'],
            trigger.data_dict['display_on_screen'],
            b'\x00\x00\x00\x00\x00',  # Unknown
            trigger.data_dict['mute_objectives'],
            trigger.data_dict['description'],
            trigger.data_dict['name'],
            trigger.data_dict['short_description'],
            len(trigger.data_dict['effects']),
            effects,
            trigger.data_dict['effect_order'],
            len(trigger.data_dict['conditions']),
            conditions,
            trigger.data_dict['condition_order'],
        ]))

    def set_name(self, val):
        self.data_dict['name'] = val + "\x00"

    def set_description(self, val):
        self.data_dict['description'] = val + "\x00"

    def set_description_stid(self, val):
        self.data_dict['description_stid'] = val

    def set_display_as_objective(self, val):
        self.data_dict['display_as_objective'] = val

    def set_short_description(self, val):
        self.data_dict['short_description'] = val + "\x00"

    def set_short_description_stid(self, val):
        self.data_dict['short_description_stid'] = val

    def set_display_on_screen(self, val):
        self.data_dict['display_on_screen'] = val

    def set_description_order(self, val):
        self.data_dict['description_order'] = val

    def set_enabled(self, val):
        self.data_dict['enabled'] = val

    def set_looping(self, val):
        self.data_dict['looping'] = val

    def set_header(self, val):
        self.data_dict['header'] = val

    def set_mute_objectives(self, val):
        self.data_dict['mute_objectives'] = val

    def set_conditions(self, val):
        self.data_dict['conditions'] = val

    def set_condition_order(self, val):
        self.data_dict['condition_order'] = val

    def set_effects(self, val):
        self.data_dict['effects'] = val

    def set_effect_order(self, val):
        self.data_dict['effect_order'] = val



import copy

from AoE2ScenarioParser.datasets import effects
from AoE2ScenarioParser.helper.retriever import find_retriever
from AoE2ScenarioParser.objects.aoe2_object import AoE2Object
from AoE2ScenarioParser.pieces.structs.effect import EffectStruct


class EffectObject(AoE2Object):
    def __init__(self,
                 effect_type,
                 ai_script_goal=-1,
                 aa_quantity=-1,
                 aa_armor_or_attack_type=-1,
                 quantity=-1,
                 tribute_list=-1,
                 diplomacy=-1,
                 number_of_units_selected=-1,
                 object_list_unit_id=-1,
                 player_source=-1,
                 player_target=-1,
                 technology=-1,
                 string_id=-1,
                 display_time=-1,
                 trigger_id=-1,
                 location_x=-1,
                 location_y=-1,
                 area_1_x=-1,
                 area_1_y=-1,
                 area_2_x=-1,
                 area_2_y=-1,
                 object_group=-1,
                 object_type=-1,
                 instruction_panel_position=-1,
                 attack_stance=-1,
                 time_unit=-1,
                 enabled_or_victory=-1,
                 food=-1,
                 wood=-1,
                 stone=-1,
                 gold=-1,
                 item_id=-1,
                 flash_object=-1,
                 force_research_technology=-1,
                 visibility_state=-1,
                 scroll=-1,
                 operation=-1,
                 object_list_unit_id_2=-1,
                 button_location=-1,
                 ai_signal_value=-1,
                 object_attributes=-1,
                 from_variable=-1,
                 variable_or_timer=-1,
                 facet=-1,
                 play_sound=-1,
                 message="",
                 sound_name="",
                 selected_object_id=-1,
                 ):
        super().__init__(locals())

    def get_content_as_string(self):
        return_string = ""

        for attribute in effects.attributes[self.data_dict['effect_type']]:
            if attribute == "effect_type" or self.data_dict[attribute] == [] or self.data_dict[attribute] == "" or \
                    self.data_dict[attribute] == " " or self.data_dict[attribute] == -1:
                continue
            return_string += "\t\t\t\t" + attribute + ": " + str(self.data_dict[attribute]) + "\n"

        return return_string

    @staticmethod
    def parse_object(parsed_data, **kwargs):  # Expected {effect=effectStruct}
        effect_struct = kwargs['effect']

        effect_type = find_retriever(effect_struct.retrievers, "Effect type").data
        parameters = effects.attributes.get(effect_type)

        parameter_dict = copy.copy(effects.empty_attributes)
        for param in parameters:
            parameter_dict[param] = find_retriever(effect_struct.retrievers, effects.attribute_naming_conversion[param]).data

        return EffectObject(
            **parameter_dict
        )

    @staticmethod
    def reconstruct_object(parsed_data, objects, **kwargs):  # Expected {effect=effect_obj, effects=effectsList}
        effect_obj = kwargs['effect']
        effects = kwargs['effects']

        data_list = list(effect_obj.data_dict.values())
        data_list.insert(1, 46)  # Check, (46)
        data_list.insert(9, -1)   # Unknown
        data_list.insert(15, -1)  # Unknown2
        data_list.insert(43, -1)  # Unknown3
        data_list.insert(48, -1)  # Unknown4

        effects.append(EffectStruct(data=data_list))

    def set_effect_type(self, val):
        self.data_dict['effect_type'] = val

    def set_ai_script_goal(self, val):
        self.data_dict['ai_script_goal'] = val

    def set_aa_quantity(self, val):
        self.data_dict['aa_quantity'] = val

    def set_aa_armor_or_attack_type(self, val):
        self.data_dict['aa_armor_or_attack_type'] = val

    def set_quantity(self, val):
        self.data_dict['quantity'] = val

    def set_tribute_list(self, val):
        self.data_dict['tribute_list'] = val

    def set_diplomacy(self, val):
        self.data_dict['diplomacy'] = val

    def set_number_of_units_selected(self, val):
        self.data_dict['number_of_units_selected'] = val

    def set_object_list_unit_id(self, val):
        self.data_dict['object_list_unit_id'] = val

    def set_player_source(self, val):
        self.data_dict['player_source'] = val

    def set_player_target(self, val):
        self.data_dict['player_target'] = val

    def set_technology(self, val):
        self.data_dict['technology'] = val

    def set_string_id(self, val):
        self.data_dict['string_id'] = val

    def set_display_time(self, val):
        self.data_dict['display_time'] = val

    def set_trigger_id(self, val):
        self.data_dict['trigger_id'] = val

    def set_location_x(self, val):
        self.data_dict['location_x'] = val

    def set_location_y(self, val):
        self.data_dict['location_y'] = val

    def set_area_1_x(self, val):
        self.data_dict['area_1_x'] = val

    def set_area_1_y(self, val):
        self.data_dict['area_1_y'] = val

    def set_area_2_x(self, val):
        self.data_dict['area_2_x'] = val

    def set_area_2_y(self, val):
        self.data_dict['area_2_y'] = val

    def set_object_group(self, val):
        self.data_dict['object_group'] = val

    def set_object_type(self, val):
        self.data_dict['object_type'] = val

    def set_instruction_panel_position(self, val):
        self.data_dict['instruction_panel_position'] = val

    def set_attack_stance(self, val):
        self.data_dict['attack_stance'] = val

    def set_time_unit(self, val):
        self.data_dict['time_unit'] = val

    def set_enabled_or_victory(self, val):
        self.data_dict['enabled_or_victory'] = val

    def set_food(self, val):
        self.data_dict['food'] = val

    def set_wood(self, val):
        self.data_dict['wood'] = val

    def set_stone(self, val):
        self.data_dict['stone'] = val

    def set_gold(self, val):
        self.data_dict['gold'] = val

    def set_item_id(self, val):
        self.data_dict['item_id'] = val

    def set_flash_object(self, val):
        self.data_dict['flash_object'] = val

    def set_force_research_technology(self, val):
        self.data_dict['force_research_technology'] = val

    def set_visibility_state(self, val):
        self.data_dict['visibility_state'] = val

    def set_scroll(self, val):
        self.data_dict['scroll'] = val

    def set_operation(self, val):
        self.data_dict['operation'] = val

    def set_object_list_unit_id_2(self, val):
        self.data_dict['object_list_unit_id_2'] = val

    def set_button_location(self, val):
        self.data_dict['button_location'] = val

    def set_ai_signal_value(self, val):
        self.data_dict['ai_signal_value'] = val

    def set_object_attributes(self, val):
        self.data_dict['object_attributes'] = val

    def set_from_variable(self, val):
        self.data_dict['from_variable'] = val

    def set_variable_or_timer(self, val):
        self.data_dict['variable_or_timer'] = val

    def set_facet(self, val):
        self.data_dict['facet'] = val

    def set_play_sound(self, val):
        self.data_dict['play_sound'] = val

    def set_message(self, val):
        self.data_dict['message'] = val

    def set_sound_name(self, val):
        self.data_dict['sound_name'] = val

    def set_selected_object_id(self, val):
        self.data_dict['selected_object_id'] = val



from AoE2ScenarioParser.helper.datatype import DataType
from AoE2ScenarioParser.helper.retriever import Retriever
import AoE2ScenarioParser.pieces.structs.aoe2_struct as structs


class PlayerDiplomacyStruct(structs.Struct):
    def __init__(self, parser_obj=None, data=None):
        retrievers = [
            Retriever("Stance with each player", DataType("u32", repeat=16)),
        ]

        super().__init__("Player Diplomacy", retrievers, parser_obj, data)



from AoE2ScenarioParser.helper.generator import create_generator
import AoE2ScenarioParser.pieces.aoe2_piece as scenario_piece


class Struct(scenario_piece.ScenarioPiece):
    def __init__(self, piece_type, retrievers, parser_obj=None, data=None):
        super().__init__(piece_type, retrievers, parser_obj, data)

        if data and parser_obj:
            gen = create_generator(data, 1)
            super().set_data_from_generator(gen)

    def _entry_to_string(self, name, data, datatype):
        return "\t\t\t" + name + ": " + data + " (" + datatype + ")\n"

    def _to_string(self):
        if self.piece_type in ("Terrain", "Unit"):
            return "."  # Recommended to keep as '.' due to the amount of tiles. (Tiny map = 14400)
        else:
            return super()._to_string()



import AoE2ScenarioParser.helper.parser as parser
from AoE2ScenarioParser.helper.retriever import find_retriever


class ScenarioPiece:
    def __init__(self, piece_type, retrievers, parser_obj=None, data=None):
        self.piece_type = piece_type
        self.retrievers = retrievers
        self.parser = parser_obj
        if data:
            self.set_data(data)

    def set_data(self, data):
        saves = {}
        if len(data) == len(self.retrievers):
            for i in range(0, len(data)):
                if self.retrievers[i].set_repeat:
                    self.retrievers[i].datatype.repeat = parser.parse_repeat_string(
                        saves,
                        self.retrievers[i].set_repeat
                    )

                if self.retrievers[i].log_value:
                    print(self.retrievers[i], "was set to:", parser.vorl(data[i]))
                self.retrievers[i].set_data(data[i])

                if self.retrievers[i].save_as is not None:
                    if type(data[i]) is not list:
                        data[i] = [data[i]]
                    saves[self.retrievers[i].save_as] = parser.vorl(data[i])
        else:
            ValueError("Data list isn't the same size as the DataType list")

    def get_value(self, retriever_key):
        return find_retriever(self.retrievers, retriever_key).data

    def get_length(self):
        total_length = 0
        try:
            for i in range(0, len(self.retrievers)):
                datatype, length = parser.datatype_to_type_length(self.retrievers[i].datatype.var)

                if datatype == "struct":
                    if type(self.retrievers[i].data) == list:
                        for continues_struct in self.retrievers[i].data:
                            length += continues_struct.get_length()
                    else:
                        length = self.retrievers[i].data.get_length()
                elif datatype == "str":
                    length = len(self.retrievers[i].data)
                total_length += length
        except TypeError:
            print(self.retrievers)
        return total_length

    def set_data_from_generator(self, generator):
        if self.parser:
            for i, retriever in enumerate(self.retrievers):
                retriever.set_data(self.parser.retrieve_value(generator, retriever))

    def _entry_to_string(self, name, data, datatype):
        return "\t" + name + ": " + data + " (" + datatype + ")\n"

    def _to_string(self):
        represent = self.piece_type + ": \n"

        for i, val in enumerate(self.retrievers):
            if type(self.retrievers[i].data) is list and len(self.retrievers[i].data) > 0:
                if isinstance(self.retrievers[i].data[0], ScenarioPiece):
                    represent += "\t" + val.name + ": [\n"
                    for x in self.retrievers[i].data:
                        represent += "\t\t" + str(x)
                    represent += "\t]\n"
                else:
                    represent += self._entry_to_string(
                        val.name,
                        str(self.retrievers[i].data),
                        str(val.datatype.to_simple_string())
                    )
            else:
                if self.retrievers[i].data is not None:
                    data = self.retrievers[i].data
                else:
                    data = "<Empty>"
                represent += self._entry_to_string(val.name, str(data), str(val.datatype.to_simple_string()))

        return represent

    def __repr__(self):
        return self._to_string()

    def __str__(self):
        return self._to_string()
